from __future__ import annotations
import logging
import itertools
import networkx
from wayfarer import functions, LENGTH_FIELD, Edge
from networkx.algorithms import eulerian_path
from networkx import NetworkXNoPath, NodeNotFound


class MultipleEnds(Exception):
    pass


log = logging.getLogger("wayfarer")


def solve_shortest_path(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    start_node: str | int,
    end_node: str | int,
    with_direction_flag: bool = True,
    weight: str | None = LENGTH_FIELD,
):
    """
    Solve the shortest path between two nodes, returning a list of Edge objects.
    Set weight to the attribute name for deciding the shortest path, or to None
    to ignore any weightings (faster).
    """
    nodes = solve_shortest_path_from_nodes(net, [start_node, end_node], weight)
    return functions.get_edges_from_nodes(
        net, nodes, with_direction_flag=with_direction_flag, length_field=weight
    )


def solve_shortest_path_from_nodes(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    node_list: list[int | str],
    weight: str | None = LENGTH_FIELD,
) -> list[int | str]:
    """
    Return a list of nodes found by solving from each node in node_list to
    the next. Set weight to the attribute name for deciding the shortest path, or to None
    to ignore any weightings (faster).
    """
    nodes_in_path = []

    for start_node, end_node in functions.pairwise(node_list):
        log.debug("Solving from %s to %s", start_node, end_node)

        if start_node == end_node:
            log.debug("Same start and end node used for path: {}".format(start_node))
        else:
            try:
                nodes_in_path += networkx.shortest_path(
                    net, source=start_node, target=end_node, weight=weight
                )
            except KeyError:
                raise

    return nodes_in_path


def solve_shortest_path_from_edges(
    net: networkx.MultiGraph | networkx.MultiDiGraph, edge_id_list: list[int | str]
):
    """
    Return a path routing from edge to edge, rather than
    from node to node
    """

    log.debug(f"Edge ids used for path solve: {edge_id_list}")

    # remove any duplicates
    edge_id_list = functions.get_unique_ordered_list(edge_id_list)

    edges = []
    removed_edges = []
    previous_edge_nodes = []  # type: list[int | str]

    for edge_id in edge_id_list:
        edge = functions.get_edge_by_key(net, edge_id)

        end_nodes = [edge.start_node, edge.end_node]

        # as a first-step remove any dual edges between points if they aren't in the input edge_id_list
        # this is to handle loops in the middle of a path (with two edges connecting the same two nodes)
        # simply setting the length to 0 for one pass isn't enough, as in the next
        # solve in this loop the shorter edge is returned. If we remove the shorter edge we can
        # ensure only the required edge is returned - see test_dual_path_middle_network

        edges_between_nodes = functions.get_edges_from_nodes(
            net, end_nodes, shortest_path_only=False
        )
        for path_edge in edges_between_nodes:
            if path_edge.key not in edge_id_list:
                removed_edges.append(path_edge)
                functions.remove_edge(net, path_edge)

        if previous_edge_nodes:
            node_list = previous_edge_nodes + end_nodes
        else:
            node_list = end_nodes

        # set the edge_id in the list to have a length of 0 to ensure it is
        # part of the route if there are alternative paths

        original_length = float(edge.attributes[LENGTH_FIELD])
        edge.attributes[LENGTH_FIELD] = 0
        log.debug(
            f"Updated edge id {edge_id} to have length 0 from {original_length:.2f}"
        )

        if edge.start_node == edge.end_node:
            # self-loop
            edges.extend(functions.get_edges_from_nodes(net, end_nodes))
        else:
            try:
                nodes = solve_shortest_path_from_nodes(net, node_list)
            except NetworkXNoPath as ex:
                log.warning(f"No path found using node_list: {node_list}")
                log.warning(ex)
                raise

            edges.extend(functions.get_edges_from_nodes(net, nodes))

            # reset original length - note not in original implementation
            # ensure this is only reset after get_edges_from_nodes has been called
            edge.attributes[LENGTH_FIELD] = original_length

        if sorted(previous_edge_nodes) == sorted(end_nodes):
            # a loop of two edges - get all edges between the nodes
            loop_edges = functions.get_edges_from_nodes(
                net, [edge.start_node, edge.end_node]
            )
            if loop_edges:
                edges.extend(loop_edges)

        previous_edge_nodes = end_nodes

    # reset network by adding back removed edges
    for removed_edge in removed_edges:
        functions.add_single_edge(net, removed_edge)

    # edge ids are not unique at this step - due to the case with doubling-back see test_simple_reversed
    unique_edge_ids = functions.get_unique_ordered_list((e.key for e in edges))
    assert len(unique_edge_ids) <= len(edges)

    # the solve should have <= all edges clicked on
    assert len(edge_id_list) <= len(unique_edge_ids)
    # all edges the user clicked on should be returned in the solve
    try:
        assert set(edge_id_list).issubset(unique_edge_ids)
    except AssertionError:
        # this can occur where there are 2 edges looping on to the same node
        # see test_dual_path
        log.debug(f"edge_id_list: {edge_id_list} unique_edge_ids: {unique_edge_ids}")
        raise

    solved_edges = []

    for edge_id in unique_edge_ids:
        edge = [e for e in edges if e.key == edge_id][
            0
        ]  # get the first edge from full list of edges
        solved_edges.append(edge)

    # start_key = edge_id_list[0]
    # start_edge = functions.get_edge_by_key(net, start_key)
    # start_node = start_edge.start_node

    return find_ordered_path(solved_edges, start_node=None)


def solve_matching_path(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    start_node: str | int,
    end_node: str | int,
    distance: int = 0,
    cutoff: int = 10,
    include_key: str | int | None = None,
) -> list[Edge] | None:
    """
    Return the path between the nodes that best matches the
    distance
    If no distance is supplied the shortest path is returned

    Cut-off is the maximum number of edges to search for
    If long solves are not as expected then increase this value

    include_key can be used to only include paths that contain this edge key
    """

    paths = []

    all_shortest_paths = list(solve_all_simple_paths(net, start_node, end_node, cutoff))

    # check for self-loops if start and end nodes are the same
    if start_node == end_node:
        all_shortest_paths.append([start_node])

    for path_nodes in all_shortest_paths:
        all_paths = functions.get_all_paths_from_nodes(
            net, path_nodes, with_direction_flag=True
        )

        for path_edges in all_paths:
            if include_key and include_key not in [e.key for e in path_edges]:
                continue
            path_length = functions.get_path_length(path_edges)
            length_difference = abs(path_length - distance)
            log.debug([p.key for p in path_edges])
            log.debug(
                "Desired length: {} Solved path length: {} Difference: {}".format(
                    distance, path_length, length_difference
                )
            )
            paths.append((length_difference, path_edges, path_length))

    # take the closest matching path to the original length based on absolute difference

    if paths:
        edges = min(paths)[
            1
        ]  # min in list based on first tuple value (length_difference)
        log.debug("Desired and solved length difference: {}".format(min(paths)[0]))
    else:
        log.warning("No paths found")
        edges = None

    return edges


def solve_matching_path_from_nodes(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    node_list: list[int | str],
    distance: int,
) -> list[Edge] | None:
    """
    From a list of unordered nodes find the longest path that connects all nodes
    Then rerun the solve from the start to the end of the path getting a path
    closest to the desired distance
    """

    if not node_list:
        return None

    # solve all paths from each node to all other nodes in the list
    all_edges = []

    # get unique combinations of node pairs

    # l = [1,2,3]
    # list(itertools.combinations(l, 2))
    # [(1, 2), (1, 3), (2, 3)]

    all_node_combinations = itertools.combinations(node_list, 2)

    for n1, n2 in all_node_combinations:
        path_nodes = solve_shortest_path(net, start_node=n1, end_node=n2)
        path_edges = functions.get_edges_from_nodes(net, path_nodes)
        if path_edges:
            all_edges.append(path_edges)

    all_edges_generator = itertools.chain(*all_edges)
    # get a unique list of edges based on key
    unique_edges = list({v.key: v for v in all_edges_generator}.values())

    # TODO following will not work if there is a closed loop
    full_path = find_ordered_path(unique_edges)
    start_node, end_node = get_path_ends(full_path)

    return solve_matching_path(net, start_node, end_node, distance)


def get_path_ends(edges: list[Edge]) -> tuple[(int | str), (int | str)]:
    """
    For a list of connected edges find the (unattached) end nodes
    TODO Check if the edges form a loop

    >>> edges = [Edge(0, 1, 1, {}), Edge(1, 2, 1, {})]
    >>> get_path_ends(edges)
    (0, 2)
    """

    # first check the case where there is only a single edge
    if len(edges) == 1:
        single_edge = edges[0]
        return single_edge.start_node, single_edge.end_node

    subnet = functions.edges_to_graph(edges)

    start_edge = edges[0]
    end_edge = edges[-1]

    start_node = functions.get_unattached_node(subnet, start_edge)
    end_node = functions.get_unattached_node(subnet, end_edge)

    return start_node, end_node


def solve_all_simple_paths(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    start_node: int | str,
    end_node: int | str,
    cutoff: int = 10,
) -> list:
    """
    Find all simple paths between the two nodes on the network.
    A simple path does not have any repeated nodes
    A wrapper function for
    `all_simple_paths <https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.simple_paths.all_simple_paths.html>`_
    TODO add ``has_path`` check prior to running

    Args:
        net: The network
        start_node: The node Id of the start node
        end_node: The node Id of the end node
        cutoff: The maximum number of edges to search for before stopping the search
    Returns:
        An iterator of a list of list of nodes

    >>> edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {"LEN_": 10}), Edge(2, 1, "C", {"LEN_": 10})]
    >>> net = functions.edges_to_graph(edges)
    >>> pths = solve_all_simple_paths(net, 0, 2)
    >>> print(list(pths))
    [[0, 1, 2], [0, 1, 2]]
    """

    all_shortest_paths = []

    log.debug("Solving from %s to %s", start_node, end_node)
    try:
        all_shortest_paths = networkx.all_simple_paths(
            net, source=start_node, target=end_node, cutoff=cutoff
        )
    except NodeNotFound as ex:
        log.error(ex)  # target node 99 not in graph
        raise

    return all_shortest_paths


def solve_all_shortest_paths(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    start_node: int | str,
    end_node: int | str,
):
    """
    Find all shortest paths between the two nodes on the network.
    This includes loops and repeated nodes.
    A wrapper function for
    `all_shortest_paths <https://networkx.org/documentation/stable/reference/algorithms/generated/networkx.algorithms.shortest_paths.generic.all_shortest_paths.html>`_
    Unsure why less results are returned in the example below than when using all_simple_paths

    Args:
        net: The network
        start_node: The node Id of the start node
        end_node: The node Id of the end node
    Returns:
        An iterator of a list of list of nodes

    >>> edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {"LEN_": 10}), Edge(2, 1, "C", {"LEN_": 10})]
    >>> net = functions.edges_to_graph(edges)
    >>> pths = solve_all_shortest_paths(net, 0, 2)
    >>> print(list(pths))
    [[0, 1, 2]]

    # noqa: E501
    """

    all_shortest_paths = []

    log.debug("Solving from %s to %s", start_node, end_node)
    try:
        all_shortest_paths = networkx.all_shortest_paths(
            net, source=start_node, target=end_node
        )
    except Exception as ex:
        log.error(ex)  # errors only seem to be raised when the iterator is used
        raise

    return all_shortest_paths


def find_ordered_path(
    edges: list[Edge],
    start_node: int | str | None = None,
    with_direction_flag: bool = True,
) -> list[Edge]:
    """
    Given a collection of randomly ordered connected edges, find the full
    path, covering all edges, from one end to the other.
    A start_node can be provided to return the edges in a specific direction.
    Uses the
    `eulerian_path <https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.euler.eulerian_path.html>`_
    function from networkx
    """

    subnet = functions.edges_to_graph(edges)

    try:
        ordered_path = list(eulerian_path(subnet, source=None, keys=True))
    except networkx.exception.NetworkXError as ex:
        log.exception(ex)
        raise

    if len(ordered_path) == 0:
        raise (ValueError("No path found for the edges. Are they connected?"))

    ordered_edges = []

    # ensure the returned path follows the direction based on the start_node to end_node
    # if a start_node is provided

    if start_node:
        end_edge = ordered_path[-1]
        end_nodes = [end_edge[0], end_edge[1]]
        if start_node in end_nodes:
            ordered_path = list(reversed(ordered_path))
        else:
            start_edge = ordered_path[0]
            start_nodes = [start_edge[0], start_edge[1]]
            if start_node not in start_nodes:
                raise ValueError(
                    "The provided start_node {} was not found at the start or end of the path".format(
                        start_node
                    )
                )

    # now get the edge objects back from the edge list

    for p in ordered_path:
        edge_key = p[2]
        edge = next(e for e in edges if e.key == edge_key)
        if with_direction_flag:
            functions.add_direction_flag(p[0], p[1], edge.attributes)
        ordered_edges.append(edge)

    return ordered_edges


# if __name__ == "__main__":
#    import doctest
#    doctest.testmod()



"""
pytest -v tests/test_splitting.py
"""

import logging
import pytest
from wayfarer import routing, splitter, Edge, WITH_DIRECTION_FIELD
from wayfarer.splitter import SPLIT_KEY_SEPARATOR
from tests import networks
from networkx import NetworkXNoPath, NodeNotFound


def test_solve_all_simple_paths():
    net = networks.simple_network()
    simple_paths = routing.solve_all_simple_paths(net, 1, 5)
    assert list(simple_paths) == [[1, 2, 3, 4, 5]]


def test_solve_all_simple_paths_no_node():
    net = networks.simple_network()
    with pytest.raises(NodeNotFound):
        list(routing.solve_all_simple_paths(net, 1, 99))


def test_solve_all_simple_paths_cutoff():
    net = networks.simple_network()
    simple_paths = routing.solve_all_simple_paths(net, 1, 5, cutoff=3)
    assert list(simple_paths) == []

    net = networks.simple_network()
    simple_paths = routing.solve_all_simple_paths(net, 1, 3, cutoff=3)
    assert list(simple_paths) == [[1, 2, 3]]


def test_solve_all_shortest_paths():
    net = networks.circle_network()
    all_paths = routing.solve_all_shortest_paths(net, 1, 3)
    # print(list(all_paths))
    assert list(all_paths) == [[1, 2, 3], [1, 4, 3]]


def test_solve_all_shortest_paths_no_node():
    net = networks.circle_network()
    # it seems NetworkXNoPath is raised rather than NodeNotFound
    # when using all_shortest_paths in networkx
    # and it is only raised when the generator is turned into a list
    with pytest.raises(NetworkXNoPath):
        print(list(routing.solve_all_shortest_paths(net, 1, 99)))


def test_get_path_ends():
    edges = [Edge(0, 1, 1, {}), Edge(1, 2, 1, {})]
    ends = routing.get_path_ends(edges)
    # print(ends)
    assert ends == (0, 2)


def test_get_path_ends_single_edge():
    edges = [Edge(0, 1, 1, {})]
    ends = routing.get_path_ends(edges)
    # print(ends)
    assert ends == (0, 1)


def test_solve_shortest_path():
    net = networks.simple_network()
    edges = routing.solve_shortest_path(net, start_node=1, end_node=5)
    edge_ids = [edge.key for edge in edges]
    assert edge_ids == [1, 2, 3, 4]


def test_solve_shortest_path_no_weight():
    net = networks.simple_network()
    edges = routing.solve_shortest_path(net, start_node=1, end_node=5, weight=None)
    edge_ids = [edge.key for edge in edges]
    assert edge_ids == [1, 2, 3, 4]


def test_solve_shortest_path_directions():
    net = networks.simple_network()
    edges = routing.solve_shortest_path(net, start_node=1, end_node=5)

    edge_directions = [edge.attributes[WITH_DIRECTION_FIELD] for edge in edges]
    assert edge_directions == [True, True, True, True]

    edges = routing.solve_shortest_path(net, start_node=5, end_node=1)

    edge_directions = [edge.attributes[WITH_DIRECTION_FIELD] for edge in edges]
    assert edge_directions == [False, False, False, False]


def test_solve_shortest_path_split_network():
    net = networks.simple_network()

    splitter.split_network_edge(net, 3, [2, 8])
    edges = routing.solve_shortest_path(net, start_node=1, end_node=5)

    # for edge in edges:
    #    print(edge)

    edge_ids = [edge.key for edge in edges]
    # print(edge_ids)
    assert edge_ids == [
        1,
        2,
        f"3{SPLIT_KEY_SEPARATOR}2",
        f"3{SPLIT_KEY_SEPARATOR}8",
        f"3{SPLIT_KEY_SEPARATOR}10",
        4,
    ]


def test_doctest():
    import doctest

    print(doctest.testmod(routing))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # test_doctest()
    # test_solve_all_simple_paths_no_node()
    # test_solve_all_simple_paths_cutoff()
    # test_solve_all_shortest_paths()
    # test_solve_all_shortest_paths_no_node()
    # test_get_path_ends()
    # test_get_path_ends_single_edge()
    # test_solve_shortest_path_directions()
    # test_solve_shortest_path_split_network()
    test_solve_shortest_path_no_weight()
    print("Done!")



"""
A collection of functions to help validate a network
"""

from __future__ import annotations
from collections import Counter
import networkx


def duplicate_keys(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
) -> list[int | str]:
    """
    Find any duplicate keys in the network
    Keys must be unique for routing to function correctly

    """

    keys = []

    for u, v, k in net.edges(keys=True):
        keys.append(k)

    duplicates = [item for item, count in Counter(keys).items() if count > 1]
    return duplicates


def valid_reverse_lookup(net: networkx.MultiGraph | networkx.MultiDiGraph):
    """
    Check if the reverse lookup dictionary
    has the same count as edges in the network
    """
    return len(net.graph["keys"]) == len(net.edges())


def recalculate_keys(net: networkx.MultiGraph | networkx.MultiDiGraph):
    """
    Recalculate keys in the reverse lookup dictionary
    These can become out-of-sync following a merge of networks e.g. with
    the compose_all function
    """

    net.graph["keys"] = {}
    for start_node, end_node, key in net.edges(keys=True):
        net.graph["keys"][key] = (start_node, end_node)


def edge_attributes(
    net: networkx.MultiGraph | networkx.MultiDiGraph, with_sample_data: bool = False
) -> list:
    """
    Return the list of attributes for the edges
    in the network
    """

    attributes_with_data = []
    attributes = []

    for u, v, d in net.edges(data=True):
        keys = sorted(d.keys())
        if keys not in attributes:
            attributes.append(keys)
            if with_sample_data:
                attributes_with_data.append(d)

    if with_sample_data:
        return attributes_with_data
    else:
        return attributes



from __future__ import annotations
from typing import NamedTuple


__version__ = "0.11.0"

LENGTH_FIELD = "LEN_"
EDGE_ID_FIELD = "EDGE_ID"
OFFSET_FIELD = "OFFSET"

NODEID_FROM_FIELD = "NODEID_FROM"
NODEID_TO_FIELD = "NODEID_TO"

WITH_DIRECTION_FIELD = "WITH_DIRECTION"

GEOMETRY_FIELD = "geometry"

SPLIT_POINT_ID_FIELD = "POINT_ID"
SPLIT_MEASURE_FIELD = "MEASURE"

SPLIT_FLAG = "IS_SPLIT"

# the Edge class

# https://mypy.readthedocs.io/en/stable/kinds_of_types.html#named-tuples


class Edge(NamedTuple):
    """
    A class representing a network edge
    """

    start_node: int | str
    end_node: int | str
    key: int | str
    attributes: dict


def to_edge(edge: tuple) -> Edge:
    """
    Convert a tuple into an Edge namedtuple

    >>> to_edge((0, 1, 1, {"LEN_": 10}))
    Edge(start_node=0, end_node=1, key=1, attributes={'LEN_': 10})

    >>> to_edge((0, 1, 1))
    Edge(start_node=0, end_node=1, key=1, attributes={})
    """
    assert len(edge) >= 3  # edges must have at least a key

    # attributes will be None if tuple is 3 in length
    if len(edge) == 3:
        if type(edge[2]) is dict:
            # ensure we don't allow edges which have data but are missing a key
            # which can happen when selecting edges using ``edges(data=True, keys=False)``
            raise ValueError(f"The edge tuple {edge} is missing a key value")
        attributes = {}
    else:
        attributes = edge[3]

    return Edge(
        start_node=edge[0],
        end_node=edge[1],
        key=edge[2],
        attributes=attributes,
    )


def to_edges(edges: tuple) -> list[Edge]:
    """
    Convert a list of tuples to a list of Edges

    >>> tuples = [(0, 1, 1, {"LEN_": 10}), (1, 2, 2, {"LEN_": 20})]
    >>> to_edges(tuples)
    [Edge(start_node=0, end_node=1, key=1, attributes={'LEN_': 10}), \
Edge(start_node=1, end_node=2, key=2, attributes={'LEN_': 20})]
    """
    return [to_edge(edge) for edge in edges]


# if __name__ == "__main__":
#    import doctest

#    doctest.testmod()
#    print("Done!")



r"""
This module contains functions relating to linear referencing.
Requires the Shapely module

Reference links:

+ https://pypi.org/project/ligeos/
+ https://www.gaia-gis.it/fossil/libspatialite/wiki?name=VirtualRouting
+ http://plugins.qgis.org/plugins/lrs/
+ http://blazek.github.io/lrs/release/help.1.2.0/index.html
+ `ogr2ogr_ZMs <https://gist.github.com/geographika/3ad5fe0c5459f1dae8d28a0ffc4b8459>`_

"""

import logging
import sys
from math import sqrt
import itertools
from itertools import islice as slice
from shapely.geometry import LineString, Point


log = logging.getLogger("wayfarer")


class IdenticalMeasuresError(Exception):
    """
    A custom exception to indicate the start and end measures are identical
    """

    pass


def check_valid_m(m_value: float, length: float, tolerance: float = 0.00001) -> bool:
    """
    Check that an m_value is a valid measure for the input length

    Args:
        m_value: The measure
        length: The length of the line
        tolerance: The tolerance the m-value can be beyond the length of the line
    Returns:
        If the line is valid

    >>> check_valid_m(m_value=10, length=20)
    True

    >>> check_valid_m(m_value=-0.213, length=20)
    Traceback (most recent call last):
    ...
    ValueError: The m value (-0.213) must be a positive number
    """

    if m_value < 0:
        raise ValueError(f"The m value ({m_value}) must be a positive number")

    if m_value > length + tolerance:
        raise ValueError(
            f"The m value ({m_value}) is higher than the line length ({length}), which is not supported"
        )

    return True


def create_line(
    line: LineString, start_m: float, end_m: float, tolerance: float = 0.00001
) -> LineString:
    """
    Create a new line feature based on a start and end measure
    This function also handles ring LineStrings (where start and end nodes are the same)

    Args:
        line: The original LineString
        start_m: The start measure to split the line
        end_m: The end measure to split the line
        tolerance: The tolerance the m-value can be beyond the length of the line
    Returns:
        A new line based on the original geometry between the two measures

    >>> ls = LineString([(0, 0), (0, 100)])
    >>> create_line(ls, start_m=0, end_m=50)
    <LINESTRING (0 0, 0 50)>
    """

    # convert any decimals to floats otherwise the comparison operators silently fail/are incorrect
    start_m, end_m = float(start_m), float(end_m)

    check_valid_m(start_m, line.length, tolerance)
    check_valid_m(end_m, line.length, tolerance)

    if start_m > end_m:
        raise ValueError(
            f"The start measure {start_m} must be less than the end measure {end_m}"
        )

    if start_m == end_m:
        raise IdenticalMeasuresError(
            f"The start measure {start_m} cannot be equal to the end measure {end_m}"
        )

    start_pt = None
    end_pt = None
    new_coords = []
    new_line = None

    for p in list(line.coords):
        # loop through all the coordinates of the original line
        distance_along = line.project(Point(p))

        if distance_along >= start_m:
            if start_pt is None:
                # create the start point
                start_pt = line.interpolate(start_m)
                new_coords.append(start_pt.coords[0])

            if distance_along < end_m:
                # add in the point just after the start point
                if distance_along != start_m:
                    # the point is not the start_pt
                    new_coords.append(p)
                else:
                    pass
                    # this is the same as the start point, unless the original feature
                    # is a ring, and needs to be closed
                    # if line.is_ring:
                    #    new_coords.append(p)

        if distance_along >= end_m:
            # create the end point, and then exit the loop
            end_pt = line.interpolate(end_m)
            new_coords.append(end_pt.coords[0])
            break

    # add an end point if the feature is a ring, and the to point
    # has not yet been added
    if line.is_ring and end_pt is None:
        end_pt = line.interpolate(end_m)
        new_coords.append(end_pt.coords[0])

    new_line = LineString(new_coords)

    return new_line


def find_common_vertices(line1: LineString, line2: LineString) -> list[Point]:
    """
    Find all the points common to both lines and return a list of touching points

    Args:
        line1: The first line
        line1: The second line
    Returns:
        A list of points shared between the lines

    >>> ls1 = LineString([(0, 0), (0, 100)])
    >>> ls2 = LineString([(0, 100), (0, 200)])
    >>> find_common_vertices(ls1, ls2)
    [<POINT (0 100)>]
    """
    touching_points = []

    # as one point will be common to both lines we only need to loop through
    # one feature's points
    points = [Point(p) for p in line2.coords]  # convert coords to Point feature

    for p in points:
        if p.intersects(line1):
            touching_points.append(p)

    return touching_points


def snap_to_ends(
    line: LineString, start_m: float, end_m: float, tolerance: float = 5
) -> tuple[float, float]:
    """
    If the measure values are within the set tolerance to the
    ends of a segment, then "snap" these values to the end.

    Args:
        line: The LineString
        start_m: The start measure
        end_m: The end measure
        tolerance: The tolerance the m-value can be beyond the length of the line
    Returns:
        A tuple of the new snapped measures

    >>> ls = LineString([(0, 0), (0, 100)])
    >>> snap_to_ends(ls, start_m=4, end_m=96, tolerance=5)
    (0, 100.0)
    >>> snap_to_ends(ls, start_m=4, end_m=96, tolerance=4)
    (4, 96)
    """

    check_valid_m(start_m, line.length, tolerance)
    check_valid_m(end_m, line.length, tolerance)

    if start_m > end_m:
        raise ValueError(
            f"The start measure {start_m} must be less than the end measure {end_m}"
        )

    # check if the measure is near the end of the line
    if (line.length - end_m) < tolerance:
        end_m = (
            line.length
        )  # line lengths can differ depending on the software so add a tolerance for any floating-point issues

    # check if the line is near the start of the line
    if (start_m) < tolerance:
        start_m = 0

    return start_m, end_m


def is_partial(
    line: LineString, start_m: float, end_m: float, tolerance: float = 0.00001
) -> bool:
    """
    Check if the measures cover a whole LineString or only a part of it

    Args:
        line: The LineString
        start_m: The start measure
        end_m: The end measure
        tolerance: The tolerance for the measures when comparing the start and end
    Returns:
        Boolean indicating if the measures only partially cover the line

    >>> ls = LineString([(0, 0), (0, 100)])
    >>> is_partial(ls, start_m=0.5, end_m=101, tolerance=1)
    False
    """

    check_valid_m(start_m, line.length, tolerance)
    check_valid_m(end_m, line.length, tolerance)

    if abs(tolerance - start_m) > tolerance:
        return True

    if abs(line.length - end_m) > tolerance:
        return True

    return False


def get_measures(line: LineString, pt1: Point, pt2: Point) -> tuple[float, float]:
    """
    Snap two points to a line and return the measures

    Args:
        line: The LineString
        pt1: A first point
        pt2: A second point
    Returns:
        A tuple of the start and end measures

    >>> ls = LineString([(0, 0), (100, 0)])
    >>> pt1 = Point(4, 0)
    >>> pt2 = Point(96, 0)
    >>> get_measures(ls, pt1, pt2)
    (4.0, 96.0)
    """

    m1 = line.project(pt1)
    m2 = line.project(pt2)
    start_m, end_m = sorted([m1, m2])

    return (start_m, end_m)


def get_measure_on_line(line: LineString, point: Point) -> float:
    """
    Return the measure along a line of a point

    Args:
        line: The LineString
        point: The point
    Returns:
        The measure (m-value) of the point along the line

    >>> line = LineString([(0, 0), (100, 0)])
    >>> point = Point(50, 0)
    >>> get_measure_on_line(line, point)
    50.0
    """
    return line.project(point)


def get_closest_point_to_measure(
    line: LineString, points: list[Point], measure: float
) -> Point:
    """
    Returns the point closest to the start measure from the
    list of points

    Args:
        line: The LineString
        points: A list of points to check
        measure: A measure along the line
    Returns:
        The measure (m-value) of the point along the line

    >>> ls = LineString([(0, 0), (0, 50), (0, 100)])
    >>> points = [Point(0, 10), Point(0, 55)]
    >>> get_closest_point_to_measure(ls, points, measure=20)
    <POINT (0 10)>
    >>> get_closest_point_to_measure(ls, points, measure=80)
    <POINT (0 55)>
    """
    measures_with_points = {}

    for point in points:
        m = line.project(point)
        measures_with_points[m] = point

    # find the closest point to the input measure
    key = min((abs(measure - i), i) for i in measures_with_points.keys())[1]

    closest_point = measures_with_points[key]
    return closest_point


def get_end_point(
    line: LineString, point: Point, common_vertices: list[Point]
) -> Point:
    """
    Get the start or end of the line closest to the input point

    Args:
        line: The input line
        point: The input point
        common_vertices: Points shared between the two lines
    Returns:
        The end point

    >>> ls = LineString([(0, 0), (100, 0)])
    >>> point = Point(30, 0)
    >>> common_vertices = [Point(0, 0), Point(70, 0)]
    >>> get_end_point(ls, point, common_vertices)
    <POINT (0 0)>
    """

    if len(common_vertices) == 2:
        start_m = line.project(point)
        end_point = get_closest_point_to_measure(line, common_vertices, start_m)
    elif len(common_vertices) == 1:
        # the line is a ring
        end_point = common_vertices[0]
    else:
        raise ValueError(
            f"Connected lines in the network should only touch 1 or 2 times (touches {len(common_vertices)} times)"
        )
    return end_point


def magnitude(pt1: Point, pt2: Point) -> float:
    """
    Return the magnitude (distance) between two points

    Args:
        pt1: The first point
        pt2: The second point
    Returns:
        The distance between the two points

    >>> pt1 = Point(0, 0)
    >>> pt2 = Point(100, 100)
    >>> round(magnitude(pt1, pt2))
    141
    """
    vect_x = pt2.x - pt1.x
    vect_y = pt2.y - pt1.y
    return sqrt(vect_x**2 + vect_y**2)


def remove_duplicates_in_line(line: LineString):
    """
    Remove any duplicate coordinates from a LineString
    Note this does not work with MultiLineStrings

    Args:
        line: The LineString
    Returns:
        A new LineString with any duplicate coordinates removed

    >>> ls = LineString([(0, 0), (1, 1), (1,1)])
    >>> remove_duplicates_in_line(ls)
    <LINESTRING (0 0, 1 1)>
    """

    if line.geom_type != "LineString":
        raise NotImplementedError(f"Geometry type {line.geom_type} is not valid")

    clean_coords = [k for k, g in itertools.groupby(line.coords)]
    return LineString(clean_coords)


def intersect_point_to_line(
    point: Point, start_point: Point, end_point: Point
) -> Point:
    """
    Get the closest point to the input point, on a straight line
    created between start point and the end point

    >>> point = Point(1, 60)
    >>> start_point = Point(0, 0)
    >>> end_point = Point(0, 100)
    >>> intersect_point_to_line(point, start_point, end_point)
    <POINT (0 60)>

    Args:
        point: The input point to place along a line
        start_point: The start point of a straight line
        end_point: The end point of a straight line
    Returns:
        A point along the straight line, or the closest end point if beyond the start and end
        points of the line

    The function will fail if two points have the same coordinates (as magnitude is 0)
    with ``ZeroDivisionError: float division``

    From the `Shapely Manual <https://shapely.readthedocs.io/en/stable/manual.html>`_ the definition of a LineString is
    "Repeated points in the ordered sequence are allowed, but may incur performance penalties and should be avoided.
    A LineString may cross itself (i.e. be complex and not simple)."
    """

    line_magnitude = magnitude(start_point, end_point)

    try:
        assert line_magnitude != 0
    except AssertionError:
        log.error("Magnitude is 0 between the two points")
        raise

    # calculate the position of a point on a line segment in 2D space
    u = (
        (point.x - start_point.x) * (end_point.x - start_point.x)
        + (point.y - start_point.y) * (end_point.y - start_point.y)
    ) / (line_magnitude**2)

    # if the closest point does not fall along the line
    # then return whichever of the end points is closest to the point

    if u < 0.00001 or u > 1:
        ix = magnitude(point, start_point)
        iy = magnitude(point, end_point)
        if ix > iy:
            return end_point
        else:
            return start_point
    else:
        ix = start_point.x + u * (end_point.x - start_point.x)
        iy = start_point.y + u * (end_point.y - start_point.y)
        return Point([ix, iy])


def get_nearest_vertex(point: Point, line: LineString) -> tuple[Point, float]:
    """
    Get the point on the line that is closest to the input point
    Based on code from `gis.stackexchange.com <http://gis.stackexchange.com/questions/396/nearest-neighbor-between-a-point-layer-and-a-line-layer>`_

    Args:
        point: The input point to place along a line
        line: The input line
    Returns:
        A tuple containing the snapped point along the line
        and its distance from the original point

    >>> point = Point(2, 21)
    >>> line = LineString([(0, 0), (0, 50), (0, 100)])
    >>> get_nearest_vertex(point, line)
    (<POINT (0 21)>, 2.0)
    """

    # remove duplicates or we will get errors with the magnitude function later on
    line = remove_duplicates_in_line(line)

    nearest_point = None
    min_dist = float(sys.maxsize)

    # loop through pairs of points
    # slice slices list but returns an iterator
    for seg_start, seg_end in zip(line.coords, slice(line.coords, 1, None)):
        line_start = Point(seg_start)
        line_end = Point(seg_end)

        intersection_point = intersect_point_to_line(point, line_start, line_end)
        cur_dist = magnitude(point, intersection_point)

        if cur_dist < min_dist:
            min_dist = cur_dist
            nearest_point = intersection_point

    log.debug(
        f"Closest point found at: {nearest_point}, with a distance of {min_dist:.2f} units."
    )

    return nearest_point, min_dist


def get_closest_line(lines: list[LineString], point: Point) -> tuple[int, Point]:
    """
    Get the closest line to a point
    """
    closest_features = []

    for code, line in lines:
        if line.geom_type != "MultiLineString":
            if line.coords[0] == line.coords[-1]:
                log.warning(
                    "Circular feature/roundabout. Removing last point to continue. "
                )
                # the magnitude of the line is messed up if try to get vertex on this
                # so try again removing last vertex
                line = LineString(line[:-1])

            closest_point, dist = get_nearest_vertex(point, line)
            closest_features.append((dist, code, closest_point))
        else:
            raise ValueError("MultiLineString comparisons not supported")

    # now we have a tuple of ids and distances, we can find the closest
    # http://stackoverflow.com/questions/644170/how-does-python-sort-a-list-of-tuples
    # if both features are equal distances, then the sorting order of the ID will take
    # precedent

    assert len(closest_features) > 0

    code = sorted(closest_features)[0][1]  # this accesses the Id
    snapped_point = sorted(closest_features)[0][2]  # this accesses the point

    return (code, snapped_point)


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)
    print("Done!")



import geojson
import math
from wayfarer import Edge, GEOMETRY_FIELD
import logging

log = logging.getLogger("wayfarer")

try:
    import fiona
    from fiona.model import Geometry
except ImportError:
    log.info("fiona is not available")

try:
    from shapely.geometry import shape
    from shapely.geometry import mapping
except ImportError:
    log.info("shapely is not available")


def edge_to_feature(edge: Edge):
    """
    Convert a wayfarer Edge to a GeoJSON feature
    """

    try:
        line = edge.attributes[GEOMETRY_FIELD]
    except KeyError:
        log.error(f"Edge: {edge}")
        log.error("Available properties: {}".format(",".join(edge.attributes.keys())))
        raise

    edge.attributes.pop(GEOMETRY_FIELD)

    # remove any Nan values that can't be converted to JSON
    for k, v in edge.attributes.items():
        if type(v) is float and math.isnan(v):
            edge.attributes[k] = None

    return geojson.Feature(
        id=edge.key, geometry=line.__geo_interface__, properties=edge.attributes
    )


def edges_to_featurecollection(edges: list[Edge]):
    """
    Convert a list of wayfarer Edges to a GeoJSON feature collection
    """

    features = [edge_to_feature(e) for e in edges]
    return geojson.FeatureCollection(features=features)


def convert_to_geojson(
    input_file: str,
    output_file: str,
    layer_name: str,
    driver_name: str = "GPKG",
    convert_multilinestring: bool = True,
) -> None:
    """
    Converts an input dataset to GeoJSON.
    Requires fiona and shapely
    Ensures each output feature has a unique id field, and that any MultiLineString features which consist
    of a single linestring are converted to linestrings.
    """

    with fiona.open(
        input_file, "r", driver=driver_name, layer=layer_name
    ) as datasource:
        geojson_schema = datasource.schema.copy()

        # set the geometry schema or we will get the following errors
        # Record's geometry type does not match collection schema's geometry type: 'LineString' != '3D MultiLineString'

        if convert_multilinestring:
            geojson_schema["geometry"] = "LineString"

        with fiona.open(
            output_file,
            "w",
            driver="GeoJSON",
            schema=geojson_schema,
            crs=datasource.crs,
            ID_GENERATE="YES",
        ) as geojson_datasource:
            # Loop through each feature in the datasource and write it to the GeoJSON file
            for feature in datasource:
                if convert_multilinestring:
                    geom = feature["geometry"]
                    if geom.type == "MultiLineString":
                        # convert to shapely geometry
                        shapely_geom = shape(geom)
                        if len(shapely_geom.geoms) == 1:
                            feature["geometry"] = Geometry.from_dict(
                                **mapping(shapely_geom.geoms[0])
                            )

                geojson_datasource.write(feature)



"""
This module handles splitting existing edges, and creating new nodes
to join the split edges
"""

from __future__ import annotations
import logging
import uuid
from collections import defaultdict, OrderedDict
from wayfarer import functions, linearref
import networkx
from wayfarer import (
    LENGTH_FIELD,
    OFFSET_FIELD,
    EDGE_ID_FIELD,
    SPLIT_FLAG,
    SPLIT_POINT_ID_FIELD,
    SPLIT_MEASURE_FIELD,
    NODEID_FROM_FIELD,
    NODEID_TO_FIELD,
    GEOMETRY_FIELD,
    Edge,
)

log = logging.getLogger("wayfarer")

SPLIT_KEY_SEPARATOR = "::"


def create_node_id(
    net: networkx.MultiGraph | networkx.MultiDiGraph, point_id: str | int
):
    # if point_id in net.nodes():
    #    raise ValueError("The point ID {} already exists as a Node Id in the network".format(point_id))
    return point_id


def create_unique_join_node() -> str:
    """
    Create a unique GUID for the join node
    """

    return str(uuid.uuid4())


def group_measures_by_edge(input):
    res = OrderedDict()

    for edge_id, measure in input:
        if edge_id in res:
            res[edge_id].append(measure)
        else:
            res[edge_id] = [measure]
    return res


def split_with_points(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    recs: list[dict],
    edge_id_field: str = EDGE_ID_FIELD,
    point_id_field: str = SPLIT_POINT_ID_FIELD,
    measure_field: str = SPLIT_MEASURE_FIELD,
):
    """
    Join a list of point records to a network, creating a new "join"
    edge

    recs is a list of dict objects that should have at a minimum the following
    fields [{"EDGE_ID": 1, "POINT_ID": 1, "MEASURE": 10.0}]
    Any additional fields will be added as properties to the new line
    """

    # first we'll group all points for each edge

    edge_splits = defaultdict(list)

    for r in recs:
        edge_id, point_id, measure = (
            r[edge_id_field],
            r[point_id_field],
            r[measure_field],
        )
        split_node = create_split_key(edge_id, measure)
        unique_edge_key = create_unique_join_node()
        from_node = create_node_id(net, point_id)

        # set the nodes as attributes so we can calculate the direction if required
        r[NODEID_FROM_FIELD] = from_node
        r[NODEID_TO_FIELD] = split_node

        # here we create a new edge joining the point to attach to the network
        # to a new node on the split line represented by the line id and measure along
        edge = Edge(
            start_node=from_node, end_node=split_node, key=unique_edge_key, attributes=r
        )

        functions.add_edge(net, **edge._asdict())  # unpack namedtuple to **kwargs
        edge_splits[edge_id].append(measure)

    for edge_id, measures in edge_splits.items():
        split_network_edge(net, edge_id, measures)


def get_split_attributes(
    original_attributes: dict,
    from_m: float | int,
    to_m: float | int,
    start_node: str | int | None = None,
    end_node: str | int | None = None,
) -> dict:
    """
    For a part of an edge that has been split
    update its attributes so the length is correct
    and any start and end nodes stored in the attributes table
    are correct

    >>> atts = {"OFFSET": 10, "LEN_": 100}
    >>> get_split_attributes(atts, 50, 70)
    {'OFFSET': 60, 'LEN_': 20, 'IS_SPLIT': True}
    """
    atts = original_attributes.copy()
    atts[LENGTH_FIELD] = to_m - from_m

    if OFFSET_FIELD in original_attributes:
        offset = original_attributes[
            OFFSET_FIELD
        ]  # when using a cumulative measure along a feature
    else:
        offset = 0

    atts[OFFSET_FIELD] = from_m + offset

    # update the nodes if provided and the fields are part of the list of attributes
    if start_node and end_node:
        if all(p in atts for p in [NODEID_FROM_FIELD, NODEID_TO_FIELD]):
            atts[NODEID_FROM_FIELD] = start_node
            atts[NODEID_TO_FIELD] = end_node

    # now add a flag to indicate this edge has been split
    atts[SPLIT_FLAG] = True

    # if geometry is being stored in the edge then create a new line
    # based on the offset
    if GEOMETRY_FIELD in atts:
        edge_id = atts[EDGE_ID_FIELD]
        ls = atts[GEOMETRY_FIELD]
        log.debug(f"Creating split geometry from {from_m} to {to_m} on {edge_id}")
        cut_line = linearref.create_line(ls, from_m, to_m)
        atts[GEOMETRY_FIELD] = cut_line

    return atts


def create_split_key(key: str | int, m: float | int, precision=3) -> str:
    """
    Function to ensure a standard format for split keys
    using the key of an edge and a measure
    E.g. "123829::154.2"
    The ``SPLIT_KEY_SEPARATOR`` can be used to find split keys
    in the network

    >>> create_split_key(123829, 154.22384972623, precision=3)
    '123829::154.224'
    >>> create_split_key(123829, 154.2)
    '123829::154.2'
    >>> create_split_key(123829, 154, precision=3)
    '123829::154'
    >>> SPLIT_KEY_SEPARATOR in "123829::154.2"
    True
    """

    # display with 3 decimal places, but remove any trailing zeros
    # and decimal place if a round number
    measure = f"{m:.{precision}f}".rstrip("0").rstrip(".")
    return f"{key}{SPLIT_KEY_SEPARATOR}{measure}"


def check_split_edges(split_edges: list[Edge]):
    # all edges should have the split flag
    is_split = [e.attributes[SPLIT_FLAG] for e in split_edges]
    assert set(is_split) == {True}

    # all edges should have the same original EdgeId
    edge_ids = [e.attributes[EDGE_ID_FIELD] for e in split_edges]
    assert len(set(edge_ids)) == 1

    return True


def get_split_nodes(split_edges: list[Edge]) -> list[int | str]:
    """
    Get the unique list of split nodes associated
    with the collection of edges

    Args:
        split_edges: a list of split edges that make up a single full edge

    Returns:
        A unique list of split nodes
    """
    nodes = []
    split_nodes = []

    for split_edge in split_edges:
        nodes.extend([split_edge.start_node, split_edge.end_node])

    for n in set(nodes):
        if SPLIT_KEY_SEPARATOR in str(n):
            split_nodes.append(n)

    return split_nodes


def get_measures_from_split_edges(split_edges: list[Edge]):
    """
    From a collection of split edges, get all the measures
    that the original edge was split
    Split nodes are in the form '2::60' (using the ``SPLIT_KEY_SEPARATOR``)

    Args:
        split_edges: a list of split edges that make up a single full edge

    Returns:
        The measures used to split the original edge
    """

    check_split_edges(split_edges)
    measures = []
    split_nodes = get_split_nodes(split_edges)

    for n in split_nodes:
        m_val = float(str(n).split(SPLIT_KEY_SEPARATOR)[1])
        measures.append(m_val)

    return measures


def unsplit_network_edges(
    net: networkx.MultiGraph | networkx.MultiDiGraph, split_edges: list[Edge]
) -> Edge:
    """
    Take a list of split network edges and join them back together
    to form the original edge. Split keys and nodes are removed from the network.
    This function can be used to then resplit the edge in different locations
    or add additional splits

    Args:
        net (object): a networkx network
        split_edges: a list of split edges that make up a single full edge

    Returns:
        The unsplit edge

    """

    check_split_edges(split_edges)

    # get all lengths
    original_length = sum([e.attributes[LENGTH_FIELD] for e in split_edges])

    # get first edge
    first_edge = min(split_edges, key=lambda e: e.attributes[OFFSET_FIELD])
    start_node = first_edge.attributes[NODEID_FROM_FIELD]
    edge_id = first_edge.attributes[EDGE_ID_FIELD]

    # get last edge
    last_edge = max(split_edges, key=lambda e: e.attributes[OFFSET_FIELD])
    end_node = last_edge.attributes[NODEID_TO_FIELD]

    # remove the split edges from the network

    for split_edge in split_edges:
        functions.remove_edge_by_key(net, split_edge.key)

    split_nodes = get_split_nodes(split_edges)

    # clean-up the split nodes - removing a node however removes
    # any connecting edges which will lead to errors, so check they are orphaned
    for n in split_nodes:
        if net.degree(n) == 0:
            net.remove_node(n)

    # copy across any attributes from the original edge
    atts = first_edge.attributes.copy()

    atts.update(
        {
            EDGE_ID_FIELD: edge_id,
            LENGTH_FIELD: original_length,
            NODEID_FROM_FIELD: start_node,
            NODEID_TO_FIELD: end_node,
        }
    )

    atts.pop(SPLIT_FLAG)
    atts.pop(OFFSET_FIELD)

    # the add_edge function takes care of adding the edge_id back to the keys
    new_edge = functions.add_edge(net, start_node, end_node, edge_id, atts)

    return new_edge


def split_network_edge(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    key: str | int,
    measures: list[int | float],
) -> list[Edge]:
    """
    Split a network edge based on a list of measures.
    We can also split a split edge but this relies on knowing the key of the
    split edge and measures would need to reflect the length of the new edge
    rather than the original edge.

    Args:
        net (object): a networkx network
        key: the key of the edge to be split
        measures: a list of measures where the line should be split

    Returns:
        A list of the new edges created by splitting
    """

    new_edges = []

    original_edge = functions.get_edge_by_key(net, key, with_data=True)

    original_length = original_edge.attributes[LENGTH_FIELD]

    validated_measures = []

    for m in measures:
        if m >= original_length:
            log.debug(
                f"Split measure {m} is greater or equal to the length {original_length} of edge {key}"
            )
        elif m <= 0:
            log.debug("Split measure is 0 or less - no need to split!")
        else:
            validated_measures.append(m)

    if len(validated_measures) == 0:
        log.debug("No measures along line - returning original edges")
        return [original_edge]

    log.debug(f"Splitting {original_edge.key} with {len(measures)} points")

    functions.remove_edge_by_key(net, original_edge.key)
    prev_node = original_edge.start_node
    from_m = 0  # type: (int | float)

    for to_m in sorted(set(validated_measures)):
        split_key = create_split_key(key, to_m)
        atts = get_split_attributes(
            original_edge.attributes, from_m, to_m, prev_node, split_key
        )
        new_edge = functions.add_edge(net, prev_node, split_key, split_key, atts)
        new_edges.append(new_edge)
        from_m, prev_node = to_m, split_key

    # add final part
    atts = get_split_attributes(
        original_edge.attributes,
        from_m,
        original_length,
        prev_node,
        original_edge.end_node,
    )

    split_key = create_split_key(key, original_length)
    new_edge = functions.add_edge(
        net, prev_node, original_edge.end_node, split_key, atts
    )
    new_edges.append(new_edge)

    return new_edges


def get_measure_for_point(line, pt):
    snapped_input_point, dist = linearref.get_nearest_vertex(pt, line)
    log.debug(f"Input point {dist:.5f} from line")
    measure = linearref.get_measure_on_line(line, snapped_input_point)
    return measure


def get_split_node_for_measure(network_edge, length, measure):
    if abs(length - measure) < 0.001:
        # don't split at the end of the line
        node_id = network_edge.end_node
    elif abs(measure - 0) < 0.001:
        # don't split at the start of the line
        node_id = network_edge.start_node
    else:
        # a new split node is created
        node_id = create_split_key(network_edge.key, measure)

    return node_id



import wayfarer
from wayfarer import loader, functions, Edge
import networkx
import pytest
import logging
import tempfile


def test_add_edge():

    net = networkx.MultiGraph()
    key = loader.add_edge(net, {"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 1})
    assert key == 1


def test_add_edge_missing_key():

    net = networkx.MultiGraph()
    with pytest.raises(KeyError):
        loader.add_edge(net, {"XEDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 1})


def test_add_edge_with_reverse_lookup():

    reverse_lookup = loader.UniqueDict()
    net = networkx.MultiGraph(keys=reverse_lookup)
    key = loader.add_edge(net, {"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 1})
    assert key == 1
    assert reverse_lookup[1] == (1, 1)


def test_create_graph():

    net = loader.create_graph()
    assert type(net) is networkx.classes.multigraph.MultiGraph
    version = wayfarer.__version__
    assert (
        str(net)
        == f"MultiGraph named 'Wayfarer Generated MultiGraph (version {version})' with 0 nodes and 0 edges"
    )
    assert net.graph["keys"] == {}


def test_create_multidigraph():

    net = loader.create_graph(graph_type=networkx.MultiDiGraph)
    assert type(net) is networkx.classes.multidigraph.MultiDiGraph
    version = wayfarer.__version__
    assert (
        str(net)
        == f"MultiDiGraph named 'Wayfarer Generated MultiDiGraph (version {version})' with 0 nodes and 0 edges"
    )


def test_create_graph_no_lookup():

    net = loader.create_graph(use_reverse_lookup=False)
    assert "keys" not in net.graph


def test_load_network_from_records():

    rec1 = {"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 2, "LEN_": 5, "PROP1": "A"}
    rec2 = {"EDGE_ID": 2, "NODEID_FROM": 2, "NODEID_TO": 3, "LEN_": 15, "PROP1": "B"}
    recs = [rec1, rec2]

    net = loader.load_network_from_records(recs)
    assert net.graph["keys"] == {1: (1, 2), 2: (2, 3)}


def test_load_network_from_records_missing_key():

    rec1 = {"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 2, "LEN_": 5, "PROP1": "A"}
    rec2 = {"XEDGE_ID": 2, "NODEID_FROM": 2, "NODEID_TO": 3, "LEN_": 15, "PROP1": "B"}
    recs = [rec1, rec2]

    with pytest.raises(KeyError):
        loader.load_network_from_records(recs)


def test_load_network_from_geometries():

    rec1 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 0), (0, 1)]},
        "properties": {"EDGE_ID": 1, "LEN_": 1},
    }
    rec2 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 1), (0, 2)]},
        "properties": {"EDGE_ID": 2, "LEN_": 1},
    }
    recs = [rec1, rec2]

    net = loader.load_network_from_geometries(recs)
    assert net.graph["keys"] == {1: ("0|0", "0|1"), 2: ("0|1", "0|2")}

    edge = functions.get_edge_by_key(net, 2)
    assert edge == Edge(
        start_node="0|1",
        end_node="0|2",
        key=2,
        attributes={
            "EDGE_ID": 2,
            "LEN_": 1,
            "NODEID_FROM": "0|1",
            "NODEID_TO": "0|2",
        },
    )


def test_load_network_from_geometries_calculated_length():

    rec1 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 0), (0, 1)]},
        "properties": {"EDGE_ID": 1},
    }
    rec2 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 1), (0, 2)]},
        "properties": {"EDGE_ID": 2},
    }
    recs = [rec1, rec2]

    net = loader.load_network_from_geometries(recs)
    assert net.graph["keys"] == {1: ("0|0", "0|1"), 2: ("0|1", "0|2")}

    edge = functions.get_edge_by_key(net, 2)
    assert edge == Edge(
        start_node="0|1",
        end_node="0|2",
        key=2,
        attributes={
            "EDGE_ID": 2,
            "LEN_": 1,
            "NODEID_FROM": "0|1",
            "NODEID_TO": "0|2",
        },
    )


def test_load_network_from_invalid_geometries():

    rec1 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 0), (0, 1)]},
        "properties": {"EDGE_ID": 1},
    }
    rec2 = {
        "geometry": {"type": "MultiLineString", "coordinates": [[(0, 1), (0, 2)]]},
        "properties": {"EDGE_ID": 2},
    }
    recs = [rec1, rec2]

    with pytest.raises(ValueError):
        loader.load_network_from_geometries(recs)


def test_load_network_from_invalid_geometries_skip():

    rec1 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 0), (0, 1)]},
        "properties": {"EDGE_ID": 1},
    }
    rec2 = {
        "geometry": {"type": "MultiLineString", "coordinates": [[(0, 1), (0, 2)]]},
        "properties": {"EDGE_ID": 2},
    }
    recs = [rec1, rec2]

    net = loader.load_network_from_geometries(recs, skip_errors=True)
    assert len(net.edges()) == 1


def test_load_network_from_invalid_geometries_missing_key():

    rec1 = {
        "geometry": {"type": "LineString", "coordinates": [(0, 0), (0, 1)]},
        "properties": {"MYKEY": 2},
    }

    recs = [rec1]

    with pytest.raises(KeyError):
        loader.load_network_from_geometries(recs)


def test_uniquedict():

    ud = loader.UniqueDict()
    ud[1] = "foo"
    ud[2] = "bar"

    assert ud == {1: "foo", 2: "bar"}


def test_uniquedict_error():

    ud = loader.UniqueDict()
    ud[1] = "foo"

    with pytest.raises(KeyError):
        ud[1] = "bar"


def test_pickling():

    net = loader.create_graph()
    loader.add_edge(net, {"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 1})
    assert net.graph["keys"][1] == (1, 1)
    assert len(net.edges()) == 1

    f = tempfile.NamedTemporaryFile(delete=False)
    fn = f.name

    loader.save_network_to_file(net, fn)

    net2 = loader.load_network_from_file(fn)
    assert net2.graph["keys"][1] == (1, 1)
    assert len(net2.edges()) == 1


def test_doctest():
    import doctest

    print(doctest.testmod(loader))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    test_doctest()
    test_add_edge()
    test_add_edge_missing_key()
    test_add_edge_with_reverse_lookup()
    test_create_graph()
    test_create_multidigraph()
    test_create_graph_no_lookup()
    test_load_network_from_records()
    test_load_network_from_records_missing_key()
    test_load_network_from_geometries()
    test_load_network_from_geometries_calculated_length()
    test_load_network_from_invalid_geometries()
    test_load_network_from_invalid_geometries_skip()
    test_load_network_from_invalid_geometries_missing_key()
    test_uniquedict()
    test_uniquedict_error()
    test_pickling()
    print("Done!")



"""
Module containing functions relating to loops
"""

from __future__ import annotations
import networkx
from networkx import cycles
from wayfarer import Edge, functions
import logging
from typing import Iterable


log = logging.getLogger("wayfarer")


def is_loop(net: networkx.MultiGraph | networkx.MultiDiGraph) -> bool:
    """
    Check if the entire network is a loop
    """

    try:
        edges = networkx.algorithms.cycles.find_cycle(net)
        if len(edges) == len(net.edges()):
            return True
        else:
            return False
    except networkx.exception.NetworkXNoCycle:
        return False


def has_loop(net: networkx.MultiGraph | networkx.MultiDiGraph) -> bool:
    """
    Check if the network has a loop
    """

    try:
        networkx.algorithms.cycles.find_cycle(net)
        return True
    except networkx.exception.NetworkXNoCycle:
        return False


def get_unique_lists(all_lists: Iterable) -> list:
    """
    Returns a unique list of lists, and preserves order

    >>> get_unique_lists([[1,2],[2,1],[1,2]])
    [[1, 2]]
    """
    unique_sorted_lists = []
    unique_lists = []

    for sublist in all_lists:
        if sorted(sublist) not in unique_sorted_lists:
            unique_sorted_lists.append(sorted(sublist))
            unique_lists.append(sublist)

    return unique_lists


def get_root_node(outdeg: list[tuple]) -> int | str:
    """
    Find the most connected node - this will be where a loop starts and ends
    If degrees are equal the first node will always be returned

    >>> outdeg = [(0, 2), (1, 3), (2, 2)]
    >>> get_root_node(outdeg)
    1
    """
    root_node = max(outdeg, key=lambda n: n[1])[
        0
    ]  # get the id of the node with the highest degree
    return root_node


def get_loop_nodes(edges: list[Edge]) -> dict:
    """
    Return a dict of nodes in a loop, with the key as the start node of the loop
    The order of the nodes in the list is sorted by node key, otherwise this is nondeterministic
    If an edge starts and ends at the same node it is not included as a loop here

    >>> edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {}), Edge(2, 0, "C", {})]
    >>> get_loop_nodes(edges)
    {0: [0, 1, 2]}
    """

    net = functions.edges_to_graph(edges)

    # to_directed adds two directed edges between each node
    loops = (
        loop for loop in cycles.simple_cycles(net.to_directed()) if len(loop) > 2
    )  # filter out any loops between 2 connected edges

    unique_loops = get_unique_lists(loops)
    if len(unique_loops) > 0:
        log.debug(("Unique loops", unique_loops))

    loop_nodes = {}

    for node_list in unique_loops:
        outdeg = net.degree(node_list)
        root_node = get_root_node(outdeg)
        loop_nodes[root_node] = sorted(node_list)

    return loop_nodes


def find_self_loop(
    net: networkx.MultiGraph | networkx.MultiDiGraph, node_id: str | int
):
    """
    For a node_id check to see if any edges start
    and end at this node
    """
    return functions.get_edges_from_nodes(
        net, [node_id, node_id], with_direction_flag=True
    )



"""
Helper module to convert between wayfarer and osmnx networks
"""

from __future__ import annotations
import wayfarer
import networkx as nx
from shapely import wkb, LineString
import osmnx as ox
import logging
import datetime as dt


from wayfarer import (
    LENGTH_FIELD,
    NODEID_FROM_FIELD,
    NODEID_TO_FIELD,
)


log = logging.getLogger("wayfarer")


def to_osmnx(
    net: nx.MultiGraph | nx.MultiDiGraph,
    geometry_field: str = "Geometry4326",
    crs: str = "epsg:4326",
) -> nx.MultiDiGraph:
    """
    Convert a wayfarer network into a osmnx network
    default_crs in osmnx is in settings.default_crs and is epsg:4326
    """

    # for timestamp see code from https://github.com/gboeing/osmnx/blob/3822ed659f1cc9f426a990c02dd8ca6b3f4d56d7/osmnx/utils.py#L50
    template = "{:%Y-%m-%d %H:%M:%S}"
    ts = template.format(dt.datetime.now())

    metadata = {
        "created_date": ts,
        "created_with": f"OSMnx {ox.__version__}",
        "crs": crs,
    }

    G = nx.MultiDiGraph(**metadata)  # wayfarer typically uses MultiGraph networks

    for tpl in net.edges(data=True, keys=True):

        edge = wayfarer.to_edge(tpl)

        try:
            edge_geometry = edge.attributes[geometry_field]
        except KeyError:
            keys = list(edge.attributes.keys())
            log.error(
                f"The attribute '{geometry_field}' was not found in the edge attributes: {keys}"
            )
            raise

        if not type(edge_geometry) is LineString:
            edge_geometry = wkb.loads(edge_geometry)

        G.add_node(
            edge.start_node, x=edge_geometry.coords[0][0], y=edge_geometry.coords[0][1]
        )
        G.add_node(
            edge.end_node, x=edge_geometry.coords[-1][0], y=edge_geometry.coords[-1][1]
        )

        properties = {
            "osmid": edge.key,
            "name": "test",
            "highway": "highway",
            "oneway": False,
            "reversed": False,
            "length": edge.attributes[LENGTH_FIELD],
            "geometry": edge_geometry,
        }

        # merge with wayfarer fields
        edge.attributes.update(properties)
        G.add_edge(edge.start_node, edge.end_node, key=edge.key, **edge.attributes)

        # now add the reversed edge
        reversed_properties = edge.attributes.copy()
        reversed_properties.update(
            {
                "reversed": True,
                NODEID_FROM_FIELD: edge.end_node,
                NODEID_TO_FIELD: edge.start_node,
            }
        )
        key = edge.key * -1

        G.add_edge(edge.end_node, edge.start_node, key=key, **reversed_properties)

    return G



"""
This module contains functions related to networks and routing
"""

from __future__ import annotations
import itertools
import copy
import logging
from collections import OrderedDict
import networkx
from wayfarer import loops
from typing import Iterable


from wayfarer import (
    EDGE_ID_FIELD,
    LENGTH_FIELD,
    NODEID_FROM_FIELD,
    NODEID_TO_FIELD,
    WITH_DIRECTION_FIELD,
    OFFSET_FIELD,
    Edge,
)


log = logging.getLogger("wayfarer")


def pairwise(iterable) -> Iterable[tuple]:
    """
    Loops through the iterable, returning adjacent pairs
    each time. Useful for returning point coordinates.

    >>> list(pairwise((0, 10, 20, 30)))
    [(0, 10), (10, 20), (20, 30)]
    """
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def get_unique_ordered_list(items: Iterable) -> list:
    """
    An order-preserving function to get a unique list of edges or nodes

    >>> get_unique_ordered_list([1, 1, 2, 3, 4, 5, 6, 5])
    [1, 2, 3, 4, 5, 6]

    Args:
        items: Any iterable
    Returns:
        A list of unique values
    """
    return list(dict.fromkeys(items))


def edges_to_graph(edges: list[Edge] | list[tuple]) -> networkx.MultiGraph:
    """
    Utility function to create a new MultiGraph based on a list of edges

    Args:
        edges: A list of Edge named tuples, or simply tuples
    Returns:
        A new network

    >>> edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {})]
    >>> net = edges_to_graph(edges)
    >>> print(net)
    MultiGraph with 3 nodes and 2 edges
    """
    net = networkx.MultiGraph()

    for edge in edges:
        if len(edge) > 2:
            net.add_edge(edge[0], edge[1], key=edge[2], attributes=edge[3])
        else:
            # when no key is provided networkx uses a 0 value for the key
            net.add_edge(edge[0], edge[1])

    return net


def add_single_edge(net: networkx.MultiGraph | networkx.MultiDiGraph, edge: Edge):
    """
    Add an edge class to a network

    Args:
        net: A network
        edge: The edge to add to the network
        end_node: The end node of the edge

    >>> net = networkx.MultiGraph()
    >>> edge = Edge(start_node=1, end_node=2, key='A', attributes={'LEN_': 10})
    >>> add_single_edge(net, edge)
    Edge(start_node=1, end_node=2, key='A', attributes={'LEN_': 10})
    >>> print(net)
    MultiGraph with 2 nodes and 1 edges
    """
    return add_edge(net, **edge._asdict())


def add_edge(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    start_node: int | str,
    end_node: int | str,
    key: int | str,
    attributes: dict,
) -> Edge:
    """
    Add an edge to a network. When adding an edge to a network, nodes are automatically
    added

    Args:
        net: A network
        start_node: The start node of the edge
        end_node: The end node of the edge
        key: The unique key of the edge
        attributes: Any attributes associated with the edge

    >>> net = networkx.MultiGraph()
    >>> add_edge(net, 1, 2, "A", {"LEN_": 10})
    Edge(start_node=1, end_node=2, key='A', attributes={'LEN_': 10})
    >>> print(net)
    MultiGraph with 2 nodes and 1 edges
    """
    new_edge = Edge(start_node, end_node, key, attributes)
    net.add_edge(
        new_edge.start_node, new_edge.end_node, new_edge.key, **new_edge.attributes
    )

    if "keys" in net.graph.keys():
        # we are using a reverse lookup dict so update this also
        net.graph["keys"][key] = (start_node, end_node)

    return new_edge


def remove_edge_by_key(
    net: networkx.MultiGraph | networkx.MultiDiGraph, key: int | str
):
    """
    Remove an edge from a network using its unique key.

    Args:
        net: A network
        key: The unique key of the edge

    >>> net = networkx.MultiGraph()
    >>> add_edge(net, 1, 2, "A", {"LEN_": 10})
    Edge(start_node=1, end_node=2, key='A', attributes={'LEN_': 10})
    >>> print(len(net.edges()))
    1
    >>> remove_edge_by_key(net, "A")
    >>> print(len(net.edges()))
    0
    """
    edge = get_edge_by_key(net, key, with_data=False)
    remove_edge(net, edge)


def remove_edge(net: networkx.MultiGraph | networkx.MultiDiGraph, edge: Edge):
    net.remove_edge(edge.start_node, edge.end_node, key=edge.key)
    """
    Remove an edge from a network using its unique key.

    Args:
        net: A network
        key: The unique key of the edge

    >>> net = networkx.MultiGraph()
    >>> edge = add_edge(net, 1, 2, "A", {"LEN_": 10})
    >>> print(len(net.edges()))
    1
    >>> remove_edge(net, edge)
    >>> print(len(net.edges()))
    0
    """
    # remove from the keys dictionary if it exists
    if "keys" in net.graph.keys():
        del net.graph["keys"][edge.key]


def get_edge_by_key(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    key: int | str,
    with_data: bool = True,
) -> Edge:
    """
    Go through all edge property dicts until the relevant key is found
    Using an iterator faster than list comprehension
    See https://stackoverflow.com/a/36413569/179520 answer from networkx dev

    https://networkx.github.io/documentation/stable/reference/classes/generated/networkx.Graph.edges.html

    400ns with reverse lookup
    1.16 seconds with generators

    >>> edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {"LEN_": 10})]
    >>> net = edges_to_graph(edges)
    >>> get_edge_by_key(net, "B", with_data=True)
    Edge(start_node=1, end_node=2, key='B', attributes={'attributes': {'LEN_': 10}})
    """

    if "keys" in net.graph:
        # this uses the reverse lookup approach of storing a dict containing the keys
        # as part of the graph
        nodes = net.graph["keys"].get(key, None)
    else:
        # otherwise we loop through all edges until the key is found
        gen = ((u, v) for u, v, k in net.edges(keys=True) if k == key)
        nodes = next(gen, None)  # min 1.17s

    if nodes is None:
        raise KeyError(f"The edge with key {key} was not found in the network")

    u, v = nodes

    try:
        if with_data:
            atts = net[u][v][key]
            edge = Edge(u, v, key, atts)
        else:
            edge = Edge(u, v, key, {})
    except KeyError:
        log.warning(
            f"The edge with start_node: {u} end_node: {v} key: {key} could not be found in the network"
        )
        raise

    return edge


def get_edges_by_attribute(net, attribute_field: str, value) -> Iterable[Edge]:
    """
    Go through all edge property dicts until the relevant attribute is found.
    Edges can have different attribute dicts, so add a check that the key exists
    If the attribute_field does not exist for a particular edge then it is ignored.
    Returns an iterator
    """

    return (
        Edge(u, v, k, d)
        for u, v, k, d in net.edges(keys=True, data=True)
        if attribute_field in d and d[attribute_field] == value
    )


# def get_edge_by_keys(net, keys, key_field=EDGE_ID_FIELD):
#    #  set(a).intersection(b)
#    return (
#        (u, v) for u, v, key_value in net.edges(data=key_field) if key_value in keys
#    )


def get_all_complex_paths(net: networkx.MultiGraph | networkx.MultiDiGraph, node_list):
    """
    For a node path, find any self-looping edges on a node
    and add these combinations to the possible path list
    """

    # first find any nodes with a self-loop
    loop_nodes = []

    for n in node_list:
        if get_edges_from_node_pair(net, n, n):
            loop_nodes.append(n)

    # see https://stackoverflow.com/a/62665978/179520

    if not loop_nodes:
        paths = [node_list]
    else:
        paths = []

        for ln in range(len(loop_nodes) + 1):
            for loop_subset in itertools.combinations(loop_nodes, ln):
                paths.append(
                    sorted(
                        node_list + list(loop_subset), key=lambda i: node_list.index(i)
                    )
                )

    return paths


def get_all_paths_from_nodes(net, node_list, with_direction_flag=False):
    """
    Get all combinations of edges along a set of nodes, rather
    than just the shortest edges
    l = [("a","b"), ("c"), ("d","e")]
    [('a', 'c', 'd'), ('a', 'c', 'e'), ('b', 'c', 'd'), ('b', 'c', 'e')]

    list(itertools.combinations([[1],[1,2]], r=2))
    [([1], [1, 2])]

    Check for self loops on every node
    """

    all_paths = []

    node_paths = get_all_complex_paths(net, node_list)

    # TODO cleanup messy loop check
    for np in list(node_paths):
        if len(set(np)) > 2:
            sn, en = np[0], np[-1]
            if sn != en:
                edges = get_edges_from_node_pair(net, sn, en, hide_log_output=True)
                if edges:
                    # there is an edge between the start and end of the path
                    # possible loop
                    extra_path = list(np)
                    extra_path.append(sn)
                    node_paths.append(extra_path)

    for node_path in node_paths:
        node_pairs = list(pairwise(node_path))

        path_list = []
        for u, v in node_pairs:  # loop through all pairs and get all edges between them
            node_edges = []
            edges = get_edges_from_node_pair(net, u, v)

            if edges:
                for key, attributes in edges.items():
                    # make a copy so client programs can modify without
                    # affecting the original edge dict
                    atts_copy = copy.deepcopy(attributes)
                    edge = Edge(start_node=u, end_node=v, key=key, attributes=atts_copy)
                    if with_direction_flag:
                        add_direction_flag(
                            **edge._asdict()
                        )  # unpack namedtuple to **kwargs

                    node_edges.append(edge)

            path_list.append(node_edges)

        # use product to get every combination of edges between the nodes
        combinatations = itertools.product(*path_list)
        all_paths.append(combinatations)

    # return a flat list of paths
    return itertools.chain(*all_paths)


def get_path_length(path_edges, length_field: str = LENGTH_FIELD):
    return sum([edge.attributes[length_field] for edge in path_edges])


def get_edges_from_node_pair(
    net, start_node: int | str, end_node: int | str, hide_log_output: bool = False
) -> networkx.classes.coreviews.AtlasView | None:
    """
    Get all edges between two nodes

    This function can return multiple edges when there is more than
    one edge between two nodes

    >>> net = networkx.MultiGraph()
    >>> net.add_edges_from([(0, 1, "A", {'val': 'foo'}), (0, 1, "B", {'val': 'bar'})])
    ['A', 'B']
    >>> get_edges_from_node_pair(net, start_node=0, end_node=1)
    AtlasView({'A': {'val': 'foo'}, 'B': {'val': 'bar'}})
    """

    if start_node == end_node:  # do not warn when checking for loops
        hide_log_output = True

    try:
        edges = net[start_node][end_node]
    except KeyError:
        if hide_log_output is False:
            log.debug("Could not find edge (%s, %s)", start_node, end_node)
        edges = None

    return edges


def get_edges_from_nodes(
    net: networkx.MultiGraph | networkx.MultiDiGraph,
    node_list: list[int | str],
    with_direction_flag: bool = False,
    length_field: str | None = LENGTH_FIELD,
    return_unique: bool = True,
    shortest_path_only: bool = True,
) -> list[Edge]:
    """
    From a list of nodes, create pairs and then get all the edges between the two nodes.
    By default only the shortest edges between the edges will be returned, but setting
    ``shortest_path_only`` to ``False`` will return all edges ignoring the ``length_field``.
    Set ``with_direction_flag`` to add a new attribute to the edge to show if it is matching the
    direction of the path.
    ``return_unique`` ensures only one copy of each edge is returned in the case of loops
    """
    edge_list = []

    node_pairs = pairwise(node_list)

    for u, v in node_pairs:
        edges = get_edges_from_node_pair(net, u, v)
        node_edges = []
        if edges:
            if shortest_path_only:
                key, attributes = get_shortest_edge(edges, length_field)
                atts_copy = copy.deepcopy(attributes)
                edge = Edge(start_node=u, end_node=v, key=key, attributes=atts_copy)
                node_edges.append(edge)
            else:
                # get all edges between the nodes, ignoring the length_field
                for key, attributes in edges.items():
                    atts_copy = copy.deepcopy(attributes)
                    node_edges.append(
                        Edge(start_node=u, end_node=v, key=key, attributes=atts_copy)
                    )

            if with_direction_flag:
                for edge in node_edges:
                    add_direction_flag(
                        **edge._asdict()
                    )  # unpack namedtuple to **kwargs

            edge_list.extend(node_edges)

    if return_unique:
        return list({v.key: v for v in edge_list}.values())
    else:
        return edge_list


def get_shortest_edge(edges: dict, length_field: str | None = LENGTH_FIELD):
    """
    From a dictionary of edges, get the shortest edge
    by its length
    If two edges have identical lengths then the edge with the higher Id
    will be returned
    """

    if not edges:
        raise ValueError("The edge list is empty")

    if length_field is None:
        # simply return the first edge
        return next(iter(edges)), edges[next(iter(edges))]
    else:
        min_key = min(
            edges, key=lambda k: edges[k][length_field]
        )  # py3 can add default=None
        return min_key, edges[min_key]


def add_direction_flag(start_node, end_node, attributes, **kwargs):
    """
    kwargs so an edge with any properties can be easily passed to the function
    """

    if not attributes:
        log.warning(
            "Cannot set the direction flag without an attributes dictionary for the edge"
        )
        return

    if NODEID_FROM_FIELD not in attributes:
        log.warning(
            "{} missing in the edge attributes. Cannot set direction.".format(
                NODEID_FROM_FIELD
            )
        )
        return

    if NODEID_TO_FIELD not in attributes:
        log.warning(
            "{} missing in the edge attributes. Cannot set direction.".format(
                NODEID_TO_FIELD
            )
        )
        return

    nodeid_from = attributes[NODEID_FROM_FIELD]
    nodeid_to = attributes[NODEID_TO_FIELD]

    if (nodeid_from, nodeid_to) == (start_node, end_node):
        # with direction
        attributes[WITH_DIRECTION_FIELD] = True
    else:
        assert (nodeid_to, nodeid_from) == (start_node, end_node)
        # against direction
        attributes[WITH_DIRECTION_FIELD] = False


def get_nodes_from_edges(edges):
    """
    From a list of edges return a unique, ordered list
    of nodes along the path
    """
    all_nodes = []
    for edge in edges:
        edge_atts = edge.attributes
        nodes = [edge_atts[NODEID_FROM_FIELD], edge_atts[NODEID_TO_FIELD]]
        if edge_atts[WITH_DIRECTION_FIELD] is False:
            nodes = reversed(nodes)

        all_nodes.extend(nodes)

    # return a unique ordered list
    return list(OrderedDict.fromkeys(all_nodes))


def get_unattached_node(net, edge):
    """
    Given an edge find which end is unattached on the
    network
    If the edge is part of a loop then return the start node
    If the edge is attached at both ends then None is returned
    """

    start_node, end_node = edge[0], edge[1]

    start_degrees = net.degree(start_node)
    end_degrees = net.degree(end_node)

    if start_degrees == 1:
        unattached_node = start_node
    elif end_degrees == 1:
        unattached_node = end_node
    else:
        if loops.has_loop(net):
            log.debug("The network is a loop")
            # return arbitrary node in this case
            unattached_node = start_node
        else:
            log.debug("Both ends of the edge are attached")
            unattached_node = None

    return unattached_node


def get_sink_nodes(net):
    """
    Get all the sinks in the network.

    out_degree_iter is the number of edges pointing out of the node
    http://networkx.lanl.gov/reference/generated/networkx.DiGraph.out_degree_iter.html

    Args:
        net (object): a networkx network

    Returns:
        sinks (list): a list of sink nodes [(0,1), (1,0)]
    """
    sinks = [node for node, degree in net.out_degree() if degree == 0]
    return sinks


def get_sink_edges(net):
    """
    Get all the sinks edges in the network. These are edges which have no edges attached
    at their downstream end

    Args:
        net (object): a networkx network

    Returns:
        end_edges (dict): a dict containing sink segment codes and their associated network edge
    """
    sinks = get_sink_nodes(net)

    return get_edges(net, sinks)


def get_source_nodes(net):
    """
    Get all the source in the network.

    in_degree_iter is the number of edges pointing into of the node
    http://networkx.lanl.gov/reference/generated/networkx.DiGraph.in_degree_iter.html

    Args:
        net (object): a networkx network

    Returns:
        sinks (list): a list of sink nodes [(0,1), (1,0)]
    """
    sources = [node for node, degree in net.in_degree() if degree == 0]
    return sources


def get_source_edges(net):
    """
    Get all the source edges in the network. These are edges which have no edges attached
    at their upstream end

    Args:
        net (object): a networkx network

    Returns:
        end_edges (dict): a dict containing source segment codes and their associated network edge
    """
    sources = get_source_nodes(net)

    return get_edges(net, sources)


def get_edges(net, nodes):
    """
    Get all the edges in the network that touch the nodes in the
    nodes list. Note only works with DiGraph or MultiDiGraph.

    Args:
        net (object): a networkx network
        nodes: (list):  a list of network nodes e.g. ``[(224966, 437657), (225195, 437940)]``

    Returns:
        end_edges (dict): a dict containing segment codes and their associated network edge
    """

    end_edges = {}

    for node in nodes:
        in_edges = net.in_edges(node)

        for edge in in_edges:
            edges = get_edges_from_node_pair(net, edge[0], edge[1])
            assert len(edges) == 1
            end_edges.update(edges)

    return end_edges


def get_ordered_end_nodes(edges: list[Edge], end_nodes: list) -> list[int]:
    """
    Ensures the end nodes are returned in the same order as the input edges

    >>> edges = [Edge(1 , 2, "A", {}), Edge(1, 0, "B", {})]
    >>> nodes = [0, 1]
    >>> get_ordered_end_nodes(edges, nodes)
    [1, 0]
    """

    # create a list of nodes from edges
    edge_nodes = [(e[0], e[1]) for e in edges]
    edge_nodes_list = list(sum(edge_nodes, ()))

    # sort the end nodes by their index in this list
    end_nodes.sort(key=lambda n: edge_nodes_list.index(n))
    return end_nodes


def get_end_nodes(edges, ordered: bool = True) -> list:
    """
    Find the end nodes of a graph

    >>> edges = [Edge(0 , 1, "A", {}), Edge(1, 2, "B", {})]
    >>> get_end_nodes(edges)
    [0, 2]
    """

    net = edges_to_graph(edges)
    out_degree = dict(net.degree())

    for n in net.nodes():
        log.debug(("Node connections for {}: {}".format(n, out_degree[n])))

    end_nodes = [n for n in out_degree if out_degree[n] == 1]

    if ordered:
        end_nodes = get_ordered_end_nodes(edges, end_nodes)

    return end_nodes


def get_multiconnected_nodes(edges, connections=2):
    """
    Get all nodes which have at least n connections

    >>> edges = [Edge(0 , 1, "A", {}), Edge(1, 2, "B", {})]
    >>> get_multiconnected_nodes(edges, connections=2)
    [1]
    """

    net = edges_to_graph(edges)
    out_degree = dict(net.degree())
    multiconnected_nodes = [n for n in out_degree if out_degree[n] >= connections]
    return multiconnected_nodes


def has_overlaps(edges: list[Edge]) -> bool:
    """
    Checks if the provided list of edges has overlaps
    If an edge is split but the measures don't overlap then this will return False
    The same edge id can be found at the start and end of the sequence if there is
    no overlap of measures

    >>> edges = [Edge(0, 1, "A", {"EDGE_ID": 1, "OFFSET": 0, "LEN_": 10}), \
Edge(1, 2, "B", {"EDGE_ID": 2, "OFFSET": 0, "LEN_": 10})]
    >>> has_overlaps(edges)
    False
    """
    # get the edge id from the attributes rather than the edge key in case
    # there are split edges which will have modified keys

    sorted_edges = sorted(edges, key=lambda e: e.attributes[EDGE_ID_FIELD])
    grouped_edges = itertools.groupby(
        sorted_edges, key=lambda e: e.attributes[EDGE_ID_FIELD]
    )

    for _, edge_group in grouped_edges:
        for edge1, edge2 in itertools.combinations(edge_group, 2):
            from_m1 = edge1.attributes.get(OFFSET_FIELD, 0)
            to_m1 = from_m1 + edge1.attributes.get(LENGTH_FIELD, 0)

            from_m2 = edge2.attributes.get(OFFSET_FIELD, 0)
            to_m2 = from_m2 + edge2.attributes.get(LENGTH_FIELD, 0)

            if max(from_m1, from_m2) < min(to_m1, to_m2):
                return True

    return False



"""
pytest -v tests/test_routing_ordered_path.py
"""

import logging
import pytest
from wayfarer import Edge, routing, loops, functions
from tests import networks
import wayfarer
from networkx.exception import NetworkXError


def test_get_path_ends():

    edges = [
        Edge(0, 1, key="A", attributes={"EDGE_ID": "A", "LEN_": 100}),
        Edge(1, 2, key="B", attributes={"EDGE_ID": "B", "LEN_": 100}),
        Edge(2, 3, key="C", attributes={"EDGE_ID": "C", "LEN_": 100}),
    ]

    ordered_edges = routing.find_ordered_path(edges, with_direction_flag=False)
    start_node, end_node = routing.get_path_ends(ordered_edges)
    assert start_node == 0
    assert end_node == 3

    # assert 2, 0 == (2,4) # WARNING will always return True! As interpreted as assert (2, 0) # True


def test_get_path_ends2():
    """
    A P shape
    """

    edges = [
        Edge(0, 1, key="A", attributes={"EDGE_ID": "A", "LEN_": 100}),
        Edge(1, 2, key="B", attributes={"EDGE_ID": "B", "LEN_": 100}),
        Edge(2, 3, key="C", attributes={"EDGE_ID": "C", "LEN_": 100}),
        Edge(3, 1, key="D", attributes={"EDGE_ID": "D", "LEN_": 100}),
    ]

    ordered_edges = routing.find_ordered_path(edges, with_direction_flag=False)
    start_node, end_node = routing.get_path_ends(ordered_edges)

    assert start_node == 0
    assert end_node == 3


def test_get_path_ends3():
    """
    A P shape - inverse loop
    """

    edges = [
        Edge(0, 1, key="A", attributes={"EDGE_ID": "A", "LEN_": 100}),
        Edge(2, 1, key="B", attributes={"EDGE_ID": "B", "LEN_": 100}),
        Edge(3, 2, key="C", attributes={"EDGE_ID": "C", "LEN_": 100}),
        Edge(3, 1, key="D", attributes={"EDGE_ID": "D", "LEN_": 100}),
    ]

    ordered_edges = routing.find_ordered_path(edges, with_direction_flag=False)
    start_node, end_node = routing.get_path_ends(ordered_edges)

    assert start_node == 0
    assert end_node == 3


def test_simple_network():

    # simple case
    net = networks.simple_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loops.has_loop(net) is False
    loop_nodes = loops.get_loop_nodes(edges)
    # no loops
    assert len(loop_nodes) == 0

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1, 5]

    edges = routing.find_ordered_path(edges)
    edge_ids = [edge.key for edge in edges]
    assert edge_ids == [1, 2, 3, 4]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=5)
    edge_ids = [edge.key for edge in edges]
    assert edge_ids == [1, 2, 3, 4]

    edge_id_list = [1, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3, 4]

    edge_id_list = [4, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [4, 3, 2, 1]

    edge_id_list = [3, 4, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [4, 3, 2, 1]


def test_single_edge_loop_network():

    net = networks.single_edge_loop_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # no loops (self-loops on one node are not included)
    assert len(loop_nodes) == 0

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1]

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=3)
    assert [edge.key for edge in edges] == [1, 2]

    edges = routing.solve_shortest_path(net, start_node=3, end_node=1)
    assert [edge.key for edge in edges] == [2, 1]

    edge_id_list = [3, 2]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [3, 2]

    edge_id_list = [2]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [2]

    # TOD this throw an error - Graph has no Eulerian path
    edge_id_list = [1, 3]
    # edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    # assert [e.key for e in edges] == [1, 2, 3]


def test_reverse_network():

    net = networks.reverse_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # no loops
    assert len(loop_nodes) == 0

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [2, 4]

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3, 4]

    edges = routing.solve_shortest_path(net, start_node=4, end_node=2)
    assert [edge.key for edge in edges] == [4, 3, 2, 1]

    edge_id_list = [1, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3, 4]

    edge_id_list = [4, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [4, 3, 2, 1]

    edge_id_list = [3, 1, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3, 4]


def test_t_network():

    # simple case
    net = networks.t_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # no loops
    assert len(loop_nodes) == 0

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1, 4, 5]

    with pytest.raises(NetworkXError):
        routing.find_ordered_path(edges)

    edges = routing.solve_shortest_path(net, start_node=1, end_node=4)
    assert [edge.key for edge in edges] == [1, 2, 3]

    # throw error
    with pytest.raises(NetworkXError):
        edge_id_list = [3, 1, 4]
        edges = routing.solve_shortest_path_from_edges(net, edge_id_list)


def test_p_network():

    # simple case
    net = networks.p_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    assert loop_nodes[2] == [2, 3, 4]

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1]

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3, 4]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=3)
    assert [edge.key for edge in edges] == [1, 2]

    edge_id_list = [1, 2, 3, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3, 4]

    edge_id_list = [4, 3, 2, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [4, 3, 2, 1]

    edge_id_list = [3, 1, 2, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    # print([e.key for e in edges])
    assert [e.key for e in edges] == [2, 3, 4, 1]


def test_double_loop_network():

    net = networks.double_loop_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # loops
    assert loop_nodes[1] == [1, 2, 3]
    assert loop_nodes[4] == [4, 5, 6]

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == []

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3, 4, 5, 6, 7]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=6)
    assert [edge.key for edge in edges] == [4, 7]

    edge_id_list = [1, 2, 3, 4, 5, 6, 7]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3, 4, 5, 6, 7]

    edge_id_list = [4, 3, 2, 1, 5, 7, 6]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [3, 2, 1, 4, 5, 6, 7]

    edge_id_list = [3, 1, 6, 2, 4, 5, 7]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    # print([e.key for e in edges])
    assert [e.key for e in edges] == [3, 2, 1, 4, 5, 6, 7]


def test_circle_network():

    net = networks.circle_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # loops
    assert loop_nodes[1] == [1, 2, 3, 4]

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == []

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3, 4]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=4)
    assert [edge.key for edge in edges] == [4]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=3)
    assert [edge.key for edge in edges] == [1, 2]

    edge_id_list = [1, 2, 3, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3, 4]

    edge_id_list = [4, 3, 2, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    # print([e.key for e in edges])
    # 4, 3, 2, 1 is equally valid, but the eulerian_path returns them in this order
    assert [e.key for e in edges] == [4, 1, 2, 3]

    edge_id_list = [2, 1, 4, 3]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    # print([e.key for e in edges])
    # 2, 1, 4, 3 is equally valid, but the eulerian_path returns them in this order
    assert [e.key for e in edges] == [2, 3, 4, 1]


def test_dual_path():

    net = networks.dual_path_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # no loops (self-loops on one node are not included)
    assert len(loop_nodes) == 0

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1]

    edges = routing.find_ordered_path(edges)
    # print([e.key for e in edges])
    assert [edge.key for edge in edges] == [1, 2, 3]

    edge_id_list = [1, 2]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2]

    edge_id_list = [1, 3]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 3]

    edge_id_list = [1, 2]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2]

    edge_id_list = [1, 2, 3]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3]


def test_dual_path_middle_network():

    net = networks.dual_path_middle_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # no loops (self-loops on one node are not included)
    assert len(loop_nodes) == 0

    # throw error
    # networkx.exception.NetworkXError: Graph has no Eulerian paths
    with pytest.raises(NetworkXError):
        edges = routing.find_ordered_path(edges)

    edge_id_list = [4, 2, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [4, 2, 1]

    edge_id_list = [1, 2, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 4]

    edge_id_list = [4, 3, 1]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [4, 3, 1]

    edge_id_list = [1, 3, 4]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 3, 4]

    # throw error
    # networkx.exception.NetworkXError: Graph has no Eulerian paths
    edge_id_list = [1, 2, 3, 4]
    with pytest.raises(NetworkXError):
        edges = routing.solve_shortest_path_from_edges(net, edge_id_list)


def test_loop_middle_network():

    net = networks.loop_middle_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    assert loop_nodes[2] == [2, 3, 4]

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1, 5]

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3, 4, 5]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=5)
    assert [edge.key for edge in edges] == [1, 5]

    edges = routing.solve_shortest_path(net, start_node=1, end_node=3)
    assert [edge.key for edge in edges] == [1, 2]

    edge_id_list = [1, 2, 3, 4, 5]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    print([e.key for e in edges])
    assert [e.key for e in edges] == [1, 2, 3, 4, 5]


def test_triple_loop_network():

    net = networks.triple_loop_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    assert loop_nodes[1] == [1, 2, 3]
    assert loop_nodes[4] == [4, 5, 6]
    assert loop_nodes[7] == [7, 8, 9]

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == []

    edges = routing.find_ordered_path(edges)
    assert [edge.key for edge in edges] == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

    edges = routing.solve_shortest_path(net, start_node=5, end_node=8)
    assert [edge.key for edge in edges] == [5, 8, 9]

    edge_id_list = [2, 8, 10]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [2, 3, 4, 8, 9, 10]


def test_bottle_network():

    net = networks.bottle_network()
    edges = wayfarer.to_edges(net.edges(keys=True, data=True))

    loop_nodes = loops.get_loop_nodes(edges)
    # no loops (self-loops on one node are not included)
    assert len(loop_nodes) == 0

    end_nodes = functions.get_end_nodes(edges)
    assert end_nodes == [1]

    edges = routing.find_ordered_path(edges)
    # print([e.key for e in edges])
    assert [edge.key for edge in edges] == [1, 2, 3]

    edge_id_list = [1, 3]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 3]

    edge_id_list = [1, 3]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 3]

    edge_id_list = [1, 2]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2]

    edge_id_list = [1, 2, 3]
    edges = routing.solve_shortest_path_from_edges(net, edge_id_list)
    assert [e.key for e in edges] == [1, 2, 3]


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # test_get_path_ends()
    # test_get_path_ends2()
    # test_get_path_ends3()
    # test_simple_network()
    # test_single_edge_loop_network()
    # test_reverse_network()
    # test_t_network()
    # test_p_network()
    # test_double_loop_network()
    test_dual_path_middle_network()
    # test_circle_network()
    # test_dual_path()
    # test_loop_middle_network()
    # test_triple_loop_network()
    # test_bottle_network()
    print("Done!")
